{-# LANGUAGE ApplicativeDo #-}

module Main where

import Control.Monad
import Data.List
import Options.Applicative
import System.Exit
import System.IO

import CPUType

data Options = Options {
    mkPath :: FilePath,
    clangPath :: FilePath,
    compareTypes :: Bool,
    outputFeatures :: Bool,
    types :: [String]
}

optParser = do
    c <- switch ( short 'c'
                    <> help "Compare CPUTYPEs and output them in order from least featured to most" )
    f <- switch ( short 'f'
                    <> help "Output CPU features instead of CPUTYPE wherever applicable" )
    m <- strOption ( short 'm'
                    <> help "Path to bsd.cpu.mk file. Default: /usr/share/mk/bsd.cpu.mk"
                    <> metavar "PATH"
                    <> value "/usr/share/mk/bsd.cpu.mk")
    e <- strOption ( short 'e'
                    <> help "Path to clang executable. Default: /usr/bin/clang"
                    <> metavar "PATH"
                    <> value "/usr/bin/clang")
    ts <- many $ argument str (metavar "CPUTYPE")

    return $ Options m e c f ts

main :: IO ()
main = do
    let parserInfo = info (optParser <**> helper) (fullDesc <> header "cputype utility")
    opts <- execParser parserInfo

    when (not (compareTypes opts) &&
          not (null $ types opts) &&
          not (outputFeatures opts)
          ) $ do
        handleParseResult . Failure $ parserFailure defaultPrefs parserInfo ShowHelpText mempty
        exitWith $ ExitFailure (-1)

    let outputLn t = if outputFeatures opts
            then do
                fs <- getCPUFeaturesFromType t (mkPath opts)
                putStrLn fs
            else putStrLn t

    supportedTypes <- getCPUTypesFromMk $ mkPath opts

    case compareTypes opts of
        False -> do
            cpuTypes <- if outputFeatures opts
                        then do
                            if null $ types opts
                            then getHostCPUType opts
                            else return $ types opts
                        else getHostCPUType opts

            forM_ cpuTypes $ \cpuType -> case cpuType `elem` supportedTypes of
                True -> do
                    outputLn cpuType
                _ -> do
                    outputLn cpuType
                    hPutStr stderr $ "This type is not supported by bsd.cpu.mk!"
                    exitWith $ ExitFailure 2
            exitSuccess

        True -> do
            forM_ (types opts) $ \t -> do
                when (not $ elem t supportedTypes) $ do
                    hPutStr stderr $ "This type is not supported by bsd.cpu.mk: " ++ t
                    exitWith $ ExitFailure 2

            typesWithFeatures <- forM (nub $ types opts) $ \t -> do
                fs <- parseCPUFeatures <$> getCPUFeaturesFromType t (mkPath opts)
                return (t, sum $ map (featureFromEnum) fs)

            forM_ (sortOn snd typesWithFeatures) (outputLn . fst)
    exitSuccess

handleMaybe descr Nothing = do
    hPutStr stderr descr
    exitWith $ ExitFailure (-1)
handleMaybe _ (Just x) = return x

getHostCPUType opts = do
    mCpuType <- getCPUTypeFromClang $ clangPath opts
    cpuType <- handleMaybe "Failed to determine CPU type using clang" mCpuType
    return [cpuType]
