{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE DeriveDataTypeable #-}

module CPUType (
    getCPUTypeFromClang,
    getCPUTypesFromMk,
    getCPUFeaturesFromType,
    parseCPUFeatures,
    featureFromEnum
    ) where

import Control.Applicative
import Data.Attoparsec.Text
import Data.Bits
import Data.Char (isAlpha, isDigit, toLower)
import Data.Data
import qualified Data.ByteString as B
import qualified Data.Text as T
import Data.Either
import Data.List
import Data.Text.Encoding
import Data.Text.Encoding.Error
import System.Process
import System.Exit

data CPUFeature = NoFeature
                | MMX
                | SSE
                | SSE2
                | SSE3
                | SSSE3
                | SSE41
                | SSE42
                | AVX
                | AVX2
    deriving (Eq, Enum, Show)

getCPUTypeFromClang :: FilePath -> IO (Maybe String)
getCPUTypeFromClang clang = do
    (result, _, stderr) <- readProcessWithExitCode clang (words "-v -march=native -c -x c /dev/null -o /dev/null") ""
    return $ case result of
        ExitSuccess -> findTargetCPUValue (words stderr)
        _ -> Nothing

findTargetCPUValue [] = Nothing
findTargetCPUValue ("-target-cpu":cpu:_) = Just cpu
findTargetCPUValue (w:ws) = findTargetCPUValue ws

getCPUFeaturesFromType :: String -> FilePath -> IO String
getCPUFeaturesFromType t bsdCpuMk = do
    dropWhileEnd (=='\n') <$> readProcess "/usr/bin/make" ["-f", bsdCpuMk, "CPUTYPE=" ++ t, "-V", "MACHINE_CPU"] ""

parseCPUFeatures :: String -> [CPUFeature]
parseCPUFeatures featuresString = filter (/=NoFeature) $ map toCPUFeature $ words featuresString

getCPUTypesFromMk :: FilePath -> IO [String]
getCPUTypesFromMk bsdCpuMk = do
    bytes <- decodeUtf8With ignore <$> B.readFile bsdCpuMk

    let lines = T.lines bytes

    return $ nub $
             map T.unpack $
             concat $
             rights $
             map (parseOnly pCPUTypeValues) lines

pCPUTypeValues :: Parser [T.Text]
pCPUTypeValues = do
    concat <$> (many $ pAssignment <|> pDirective <|> pDirectiveAfterSlash)

pAssignment = do
    skipWhile isHorizontalSpace
    string "CPUTYPE"
    skipWhile isHorizontalSpace
    string "="
    skipWhile isHorizontalSpace
    (:[]) <$> takeTill isEndOfLine

pDirective = do
    char '.'
    pDirectiveBody

pDirectiveAfterSlash = do
    skipWhile isHorizontalSpace
    pDirectiveBody

pDirectiveBody = do
    many1 $ do
        skipWhile (/='$')
        string "${CPUTYPE}"
        skipWhile isHorizontalSpace
        string "==" <|> string "!="
        skipWhile isHorizontalSpace
        char '"'
        takeTill (=='"')

toCPUFeature "avx2" = AVX2
toCPUFeature "avx" = AVX
toCPUFeature "sse42" = SSE42
toCPUFeature "sse41" = SSE41
toCPUFeature "ssse3" = SSSE3
toCPUFeature "sse3" = SSE3
toCPUFeature "sse2" = SSE2
toCPUFeature "sse" = SSE
toCPUFeature "mmx" = MMX
toCPUFeature x = NoFeature

featureFromEnum :: CPUFeature -> Int
featureFromEnum NoFeature = 0
featureFromEnum x = 1 `shift` (fromEnum x - 1)
